# -*- coding: utf-8 -*-
##------------------------------------------
##   My Library YAML file process Ver 0.02
##
##               2022.01.11 Masahiro Izutsu
##------------------------------------------
## my_yaml.py          (mylib_yaml.py)
##      Ver 0.02    機能拡張      2022.09.15

import my_logging
import yaml
import codecs
import os

class YamlProcess:
    logger = None
    out_path = None
    
    # 初期化
    #   out_path:   出力ファイル・パス
    #   logsel: ログ出力選択
    def __init__(self, out_path, logsel = 4):
        self.logger = my_logging.get_module_logger_sel(__name__, logsel)
        self.out_path = out_path

    # セクション/キーの追加書き込み
    #   section:    セクション名
    #   key:        データ名
    #   data:       オブジェクト
    def write_append_key(self, section, key, data):
        if os.path.isfile(self.out_path):
            with codecs.open(self.out_path, 'r+', 'utf-8') as f:
                obj = None
                try:
                    obj = yaml.safe_load(f)
                except yaml.YAMLError as exc:
                    print(exc)

                if obj != None and section in obj:
                    obj[section][key] = data
                    f.seek(0)
                    yaml.dump(obj, f, encoding='utf-8', allow_unicode=True)                
                    self.logger.debug(yaml.dump(obj))
                else:
                    new_data = {key: data}
                    self.write_append(section, new_data)
        else:
            new_data = {key: data}
            self.write_section(section, new_data)

    # セクションの追加書き込み
    #   section:    セクション名
    #   data:       オブジェクト
    def write_append(self, section, data):
        if os.path.isfile(self.out_path):
            with codecs.open(self.out_path, 'r+', 'utf-8') as f:
                try:
                    obj = yaml.safe_load(f)
                    obj[section] = data
                    f.seek(0)
                    yaml.dump(obj, f, encoding='utf-8', allow_unicode=True)                
                    self.logger.debug(yaml.dump(obj))

                except yaml.YAMLError as exc:
                    self.logger.error(exc)

        else:
            self.write_section(section, data)

    # セクションの書き込み
    #   section:    セクション名
    #   data:       オブジェクト
    def write_section(self, section, data):
        obj = {section: data}
        self.write_object(obj)

    # オブジェクトの書き込み
    #   obj:       オブジェクト
    def write_object(self, obj):
        with codecs.open(self.out_path, 'w', 'utf-8') as f:
            yaml.dump(obj, f, encoding='utf-8', allow_unicode=True)

        self.logger.debug(yaml.dump(obj))

    # セクション/サブ・セクション/キーの読み出し
    #   section:    セクション名
    #   sub_sec:    サブ・セクション名
    #   key:        データ名
    def load_section_sub_key(self, section, sub_sec, key):
        obj = self.load_section_key(section, sub_sec)
        data = None
        if obj != None and key in obj:
            data = obj[key]
        return data

    # セクション/キーの読み出し
    #   section:    セクション名
    #   key:        データ名
    def load_section_key(self, section, key):
        obj = self.load_section(section)
        data = None
        if obj != None and key in obj:
            data = obj[key]
        return data

    # セクションの読み出し
    #   戻り値:      セクション・オブジェクト
    def load_section(self, section):
        obj = self.load_object()
        data = None
        if obj != None and section in obj:
            data = obj[section]
        return data
        
    # オブジェクトの読み出し
    #   戻り値:      オブジェクト
    def load_object(self):
        obj = None
        with codecs.open(self.out_path, 'r', 'utf-8') as f:
            try:
                obj = yaml.safe_load(f)
            except yaml.YAMLError as exc:
                self.logger.error(exc)
        return obj


#-----Test routine-----
# $ python3 my_yaml.py
#
if __name__ == "__main__":
    from os.path import expanduser

    FILE_PATH = expanduser('./test.yaml')
    MAIL_SET = 'MailServer'
    UART_SET = 'UARTport'
    JTALK_SET = 'JTalkVoice'
    KEY_ACCOUNT = 'account'
    KEY_PASSWORD = 'password'
    KEY_HOST = 'host'
    KEY_PORT = 'port'
    KEY_FROM = 'from'
    KEY_TO ='to'
    KEY_VOICE = 'voice'
    KEY_ENABLE = 'enable'
    
    logsel = 1

    # 初期設定ファイル作成
    myyaml = YamlProcess(FILE_PATH, logsel)

    # UART ポート情報を書く
    port = '/dev/ttyUSB0'
    data = {KEY_PORT : port}
    myyaml.write_append(UART_SET, data)

    # OpenJTalk 情報を書く
    voice = 0
    data = {KEY_VOICE : voice}
    myyaml.write_append(JTALK_SET, data)

    # SMTPサーバー情報を書く
    enbl = 'No'
    account = 'aipromotion999@gmail.com'
    password = 'alnxncjgkmrzxwua'
    host = 'smtp.gmail.com'
    port = 465
    from_email = 'aipromotion999@gmail.com'
    to_email = 'izutsum-sp@docomo.ne.jp'
    data = {KEY_ACCOUNT:account, KEY_PASSWORD:password, KEY_HOST:host, KEY_PORT:port, KEY_FROM:from_email, KEY_TO:to_email, KEY_ENABLE:enbl}
    myyaml.write_append(MAIL_SET, data)

    # 設定を読み出す
    def _read_section_key(section, key):
        obj = myyaml.load_section_key(section, key)
        print(f"{section} {key} : '{obj}'", type(obj))

    _read_section_key(UART_SET, KEY_PORT)
    _read_section_key(JTALK_SET, KEY_VOICE)
    _read_section_key(MAIL_SET, KEY_ACCOUNT)
    _read_section_key(MAIL_SET, KEY_PASSWORD)
    _read_section_key(MAIL_SET, KEY_HOST)
    _read_section_key(MAIL_SET, KEY_PORT)
    _read_section_key(MAIL_SET, KEY_FROM)
    _read_section_key(MAIL_SET, KEY_TO)
    _read_section_key(MAIL_SET, KEY_ENABLE)

